@@ -1237,35 +1237,36 @@ ncp_find_conn_spec(const char *server, const char *user, const char *password,
 			}
 			fclose(nwc);
 		}
+	}
 
-		if (strlen(spec.user) == 0)
+
+	if (strlen(spec.user) == 0)
+	{
+		return &spec;
+	}
+
+	if (strlen(spec.password) == 0)
+	{
+		char *password;
+		if (!(isatty(0) && isatty(1)))
 		{
-			return &spec;
+			return NULL;
 		}
+		printf("Logging into %s as %s\n",
+		       spec.server, spec.user);
 
-		if (strlen(spec.password) == 0)
+		password = getpass("Password: ");
+		if (strlen(password) > sizeof(spec.password))
 		{
-			char *password;
-			if (!(isatty(0) && isatty(1)))
-			{
-				return NULL;
-			}
-			printf("Logging into %s as %s\n",
-			       spec.server, spec.user);
-
-			password = getpass("Password: ");
-			if (strlen(password) > sizeof(spec.password))
-			{
-				return NULL;
-			}
-			strcpy(spec.password, password);
+			return NULL;
 		}
-		else
+		strcpy(spec.password, password);
+	}
+	else
+	{
+		if (strcmp(spec.password, NWC_NOPASSWORD) == 0)
 		{
-			if (strcmp(spec.password, NWC_NOPASSWORD) == 0)
-			{
-				*spec.password = '\0';
-			}
+			*spec.password = '\0';
 		}
 	}
 
@@ -1349,6 +1350,7 @@ ncp_initialize_as(struct ncp_conn *conn,
 			{
 				return -1;
 			}
+			password = NWC_NOPASSWORD;
 			continue;
 		}
 		i += 1;
@@ -2739,3 +2741,98 @@ ncp_write(struct ncp_conn *conn, const char *file_id,
 	}
 	return already_written;
 }
+
+int
+ncp_copy_file(struct ncp_conn *conn,
+	      const char source_file[6],
+	      const char target_file[6],
+	      __u32 source_offset,
+	      __u32 target_offset,
+	      __u32 count,
+	      __u32 *copied_count)
+{
+	int result;
+
+	ncp_init_request(conn);
+
+	ncp_add_byte(conn, 0);	/* reserved */
+	ncp_add_mem(conn, source_file, 6);
+	ncp_add_mem(conn, target_file, 6);
+	ncp_add_dword(conn, source_offset);
+	ncp_add_dword(conn, target_offset);
+	ncp_add_dword(conn, count);
+
+	if ((result = ncp_request(conn, 74)) != 0)
+	{
+		ncp_unlock_conn(conn);
+		return result;
+	}
+
+	*copied_count = ncp_reply_dword(conn, 0);
+	return 0;
+}
+
+int
+ncp_get_broadcast_message(struct ncp_conn *conn, char message[256])
+{
+	int result;
+	int length;
+
+	ncp_init_request_s(conn, 1);
+
+	if ((result = ncp_request(conn, 21)) != 0)
+	{
+		ncp_unlock_conn(conn);
+		return result;
+	}
+
+	length = ncp_reply_byte(conn, 0);
+	message[length] = 0;
+	memcpy(message, ncp_reply_data(conn, 1), length);
+	ncp_unlock_conn(conn);
+	return 0;
+}
+
+int
+ncp_dealloc_dir_handle(struct ncp_conn *conn, __u8 dir_handle)
+{
+	int result;
+
+	ncp_init_request_s(conn, 20);
+	ncp_add_byte(conn, dir_handle);
+
+	if ((result = ncp_request(conn, 22)) != 0)
+	{
+		ncp_unlock_conn(conn);
+		return result;
+	}
+	ncp_unlock_conn(conn);
+	return 0;
+}
+
+int
+ncp_alloc_short_dir_handle(struct ncp_conn *conn,
+			   struct nw_info_struct *dir,
+			   word   alloc_mode,
+			   byte  *target)
+{
+	int result;
+
+	ncp_init_request(conn);
+	ncp_add_byte(conn, 12);	/* subfunction */
+	ncp_add_byte(conn, 0);	/* dos name space */
+	ncp_add_byte(conn, 0);	/* reserved */
+	ncp_add_word(conn, htons(alloc_mode));
+	ncp_add_handle_path(conn, dir->volNumber, dir->DosDirNum,
+			    1, NULL);
+
+	if ((result = ncp_request(conn, 87)) != 0)
+	{
+		ncp_unlock_conn(conn);
+		return result;
+	}
+
+	*target = ncp_reply_byte(conn, 0);
+	ncp_unlock_conn(conn);
+	return result;
+}
