@@ -62,7 +62,7 @@ main(int argc, char **argv)
 	{
 		found = 1;
 		printf("%-52s", q.object_name);
-		printf("%08x\n", q.object_id);
+		printf("%08X\n", (unsigned int)q.object_id);
 	}
 
 	if ((found == 0) && (isatty(1)))
